/*
  Some basic documentation should go here

  This is a MINIMAL .cc file.
 */
#include <fwdpp/diploid.hh>
#include <config.h>
//If you make a custom mutation type,
//you need to define it above,
//and adjust this using declaration:
using mtype = KTfwd::mutation;

//At this point, you need to declare your containers, etc.
//See fwdpp's example programs, etc.

using namespace std;

int main(int argc, char ** argv)
{
}
